import time
import argparse
from .resources.local import MariTalkLocal

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Start MariTalk Local server.")
    parser.add_argument(
        "--license",
        type=str,
        required=True,
        help="License key for downloading the model.",
    )
    parser.add_argument(
        "--path", type=str, default="~/bin/maritalk", help="Path for the executable."
    )
    parser.add_argument(
        "--cuda",
        type=int,
        choices=[11, 12],
        default=None,
        help="Installed CUDA version.",
    )
    parser.add_argument("--port", type=int, default=9000, help="The HTTP port to bind.")
    args = parser.parse_args()

    client = MariTalkLocal(port=args.port)
    client.start_server(
        license=args.license,
        bin_path=args.path,
        cuda_version=args.cuda,
    )

    while True:
        output = client.process.stdout.readline()
        if client.process.poll() is not None and output == b"":
            break
        if output:
            print(output.decode(), end='', flush=True)



import os
import argparse
from .resources.local import download, check_gpu, find_libs

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Download MariTalk Local server.")
    parser.add_argument(
        "--license",
        type=str,
        required=True,
        help="License key for downloading the model.",
    )
    parser.add_argument(
        "--path",
        type=str,
        default="~/bin/maritalk",
        help="Path to save the executable.",
    )
    parser.add_argument(
        "--cuda",
        type=int,
        choices=[11, 12],
        default=None,
        help="Installed CUDA version.",
    )
    args = parser.parse_args()

    if not args.cuda:
        check_gpu()
    detected_versions = find_libs()

    dependencies = {"cuda_version": args.cuda or detected_versions["cuda_version"]}

    if dependencies["cuda_version"] is None:
        raise Exception(
            "No libcublas.so found. cuBLAS v11 or v12 is required to run MariTalk. You can manually set the version using the `cuda_version` argument."
        )

    args.path = os.path.expanduser(args.path)
    bin_folder = os.path.dirname(args.path)
    if bin_folder:
        os.makedirs(bin_folder, exist_ok=True)
    download(args.license, args.path, dependencies)



import json
import requests
from typing import AsyncGenerator, Generator, List, Dict, Union, Optional
from requests.exceptions import HTTPError
from http import HTTPStatus

import httpx


class MaritalkHTTPError(HTTPError):
    def __init__(self, request_obj):
        self.request_obj = request_obj
        try:
            response_json = request_obj.json()
            if "detail" in response_json:
                api_message = response_json["detail"]
            elif "message" in response_json:
                api_message = response_json["message"]
            else:
                api_message = response_json
        except:
            api_message = request_obj.text

        self.message = api_message
        self.status_code = request_obj.status_code

    def __str__(self):
        status_code_meaning = HTTPStatus(self.status_code).phrase
        formatted_message = f"HTTP Error: {self.status_code} - {status_code_meaning}\nDetail: {self.message}"
        return formatted_message


class MariTalk:
    def __init__(self, key: str, api_url: str = "https://chat.maritaca.ai/api", model: Optional[str] = None):
        if model is None:
            raise ValueError("You must set one of the available models: sabia-2-medium, sabia-2-small, sabia-2-medium-2024-03-13, sabia-2-small-2024-03-13 and maritalk-2024-01-08 (deprecated)")
        self.key = key
        """@private"""
        self.api_url = api_url
        """@private"""
        self.model = model
        """@private"""

    @staticmethod
    async def _async_generate(url: str, headers: dict, data: dict) -> AsyncGenerator:
        """
        Returns an async generator that yields the response from the server.
        Args:
            url (`str`): 
                The URL to connect to.
            headers (`dict`):
                The headers to be sent with the request.
            data (`dict`):
                The data to be sent with the request.
        """
        try:
            async with httpx.AsyncClient() as client:
                async with client.stream("POST", url, data=json.dumps(data), headers=headers) as response:
                    response.raise_for_status()
                    async for line in response.aiter_lines():
                        if line == "":
                            continue
                        if line.startswith("data: "):
                            data = line.replace("data: ", "")
                            if data:
                                yield json.loads(data)
        except Exception as e:
            raise Exception(f"An error occurred while trying to connect to the server: {str(e)}")

    @staticmethod
    def _generate_streaming(response: requests.Response) -> Generator:
        """
        Returns a generator that yields the response from the server.
        Args:
            response (`requests.Response`):
                The response object returned from the server.
        """
        try:
            for contents in response.iter_lines():
                payload = contents.decode("utf-8")
                if payload == "":
                    continue
                if payload.startswith("data: "):
                    data = payload.replace("data: ", "")
                    if data:
                        yield json.loads(data)
        except Exception as e:
            raise Exception(f"An error occurred while trying to connect to the server: {str(e)}") 
        

    def generate(
        self,
        messages: Union[str, List[Dict[str, str]]],
        chat_mode: bool = True,
        temperature: float = 0.7,
        top_p: float = 0.95,
        max_tokens: int = 512,
        do_sample: bool = True,
        stopping_tokens: List = [],
        stream: bool = False,
        return_async_generator: bool = False,
        num_tokens_per_message: int = 4
    ):
        """
        Generate a response from a list of messages.

        Args:
            messages (`Union[str, List[Dict[str, str]]]`, *optional*):
                If chat_mode=True, messages should be a string representing a single user message or a list of messages comprising a conversation between the user and the assistant.
                If messages is a list, each item of the list should be a dictionary containing the keys `role` and `content`. For example:
                ```
                messages = [
                    {"role": "user", "content": "bom dia, esta é a mensagem do usuario"},
                    {"role": "assistant", "content": "bom dia, esta é a resposta do assistente"},
                    {"role": "user", "content": "Você pode me falar quanto é 25 + 27?"},
                ]
                ```
                If chat_mode=False, messages should be a string representing a prompt.
            chat_mode (`bool`, *optional*, defaults to True):
                If True, the model will run in chat mode, in which messages is either a string representing a single user message or a list of messages representing the conversation between the user and the assistant.
                If False, messages should be a string representing the prompt. chat_mode=False is recommended when using few-shot examples.
            temperature (`float`, *optional*, defaults to `0.7`):
                The sampling temperature for the next token probability. Higher values generate more random texts, while lower values will make it more deterministic.
            top_p (`float`, *optional*, defaults to `0.95`):
                The top probability mass to use on nucleus sampling. Read more at: https://arxiv.org/abs/1904.09751.
            max_tokens (`int`, *optional*, defaults to `512`):
                Maximum number of tokens to generate.
            do_sample (`bool`, *optional*, defaults to `True`):
                Whether to use sampling or not. `True` value means non-deterministic generations using sampling parameters and `False` value means deterministic generation using greedy decoding.
            stopping_tokens (`List`, *optional*):
                A list of tokens to use as a stop criteria.
            stream (`bool`, *optional*, defaults to `False`):
                If True, the function will stream the response from the server. This is useful when generating a large amount of tokens, as it will allow you to consume the response as it is being generated. If False, the function will return the entire response at once.
            return_async_generator (`bool`, *optional*, defaults to `False`):
                If True, the function will return an async generator that yields the response from the server. If False, the function will return a generator. This argument is only used when `stream=True`.
            num_tokens_per_message (`int`, *optional*, defaults to `4`):
                The number of tokens to yield per message when using the async generator. This argument is only used when `stream=True`.
        """

        if chat_mode:
            if not (
                isinstance(messages, str)
                or (
                    isinstance(messages, list)
                    and len(messages) > 0
                    and isinstance(messages[0], dict)
                )
            ):
                raise TypeError(
                    "Invalid `messages` argument format. It's expected to be a `str` or a list of dictionaries containing `role` and `content` keys."
                )
        else:
            if not isinstance(messages, str):
                raise TypeError(
                    "Invalid `messages` argument format. It's expected to be a `str` when chat_mode=False."
                )

        body = {
            "model": self.model,
            "messages": messages,
            "chat_mode": chat_mode,
            "do_sample": do_sample,
            "temperature": temperature,
            "top_p": top_p,
            "max_tokens": max_tokens,
            "stopping_tokens": stopping_tokens,
            "stream": stream,
            "num_tokens_per_message": num_tokens_per_message
        }

        headers = {}

        if self.key is not None:
            headers["Authorization"] = "Key " + self.key
    
        if stream:
            if return_async_generator:
                return self._async_generate(self.api_url + "/chat/inference", headers, body)
            else:
                response = requests.post(
                    self.api_url + "/chat/inference", json=body, headers=headers, stream=True
                )
                if not response.ok:
                    raise MaritalkHTTPError(response)
                return self._generate_streaming(response)
        else:
            response = requests.post(
                self.api_url + "/chat/inference", json=body, headers=headers
            )

            if response.ok:
                return response.json()
            else:
                raise MaritalkHTTPError(response)



import os
import maritalk

client = maritalk.MariTalkLocal()
client.start_server(license=os.environ["MARITALK_LICENSE"])

# Question selected from the FaQuAD dataset:
# https://huggingface.co/datasets/eraldoluis/faquad?row=32
result = client.generate([
    {
        "role": "user",
        "content": """Com base no texto abaixo, resposta a pergunta "Qual o objetivo do POSCOMP?"
Contexto: O Exame Nacional para Ingresso na Pós-Graduação em Computação (POSCOMP) é um exame aplicado em todas as regiões do País. Em parceria com a Sociedade Peruana de Computação, desde 2006 o Exame passou a ser realizado no Peru. O POSCOMP testa conhecimentos na área de Computação e tem como objetivo específico avaliar os conhecimentos de candidatos a Programas de Pós-Graduação em Computação oferecidos no Brasil. A grande maioria dos Programas de Pós-Graduação no País utiliza, de alguma forma, o resultado do POSCOMP em seu processo seletivo."""
    }
])
print(result["output"])



"""
Benchmark tool for MariTalk Local server.

Usage example:

$ python benchmark.py \
    --concurrency 1,2,4,8 \
    --n-repeats 3 \
    --prompt-size 1_000 \
    --max-tokens 500 \
    --tokenizer maritaca-ai/maritalk-tokenizer-small

Install required dependencies:
$ pip install maritalk transformers pandas matplotlib tqdm
"""

import time
import string
import random
import argparse
import statistics
import pandas as pd
pd.set_option("display.precision", 1)
from tqdm import tqdm
import concurrent.futures
import matplotlib.pyplot as plt
from collections import defaultdict
from transformers import AutoTokenizer
import maritalk


def request(client, prompt, max_tokens):
    """
    Call MariTalk Local API
    """
    start_time = time.time()
    result = client.generate_raw(
        prompt,
        max_tokens=max_tokens,
        temperature=3.0,  # to force longer generations
    )
    total_time = time.time() - start_time
    return {
        "output": result["output"],
        "generation_time": result["generation_time"],
        "total_time": total_time,
    }


parser = argparse.ArgumentParser()
parser.add_argument(
    "--n-repeats",
    default=5,
    type=int,
    help="Number of times to repeat the request to obtain an average",
)
parser.add_argument(
    "--prompt-size", default=550, type=int, help="Input prompt size in tokens"
)
parser.add_argument(
    "--max-tokens", default=150, type=int, help="Maximum tokens to be generated"
)
parser.add_argument(
    "--tokenizer",
    type=str,
    default="maritaca-ai/maritalk-tokenizer-small",
    help="Tokenizer to be used to count tokens",
)
parser.add_argument(
    "--concurrency",
    default="1,2",
    type=str,
    help="A list of concurrent requests to perform",
)
parser.add_argument(
    "--license",
    default=None,
    required=False,
    type=str,
    help="Your MariTalk Local license to call `start_server` method. Ignore it if already started a server.",
)
args = parser.parse_args()
args.concurrency = [int(n) for n in args.concurrency.split(",")]

client = maritalk.MariTalkLocal()

# Start MariTalk Local server if a license is provided
# or use an already started server otherwise
if args.license:
    client.start_server(license=args.license)

tokenizer = AutoTokenizer.from_pretrained(args.tokenizer)
results = []

for concurrency in tqdm(args.concurrency):
    for i in range(args.n_repeats):
        prompt = " ".join([str(i) for i in range(args.prompt_size)])
        prompt_size = len(tokenizer.tokenize(prompt))
        if prompt_size > args.prompt_size:
            prompt = tokenizer.decode(
                tokenizer.encode(prompt)[: args.prompt_size],
                skip_special_tokens=True,
            )
            prompt_size = len(tokenizer.tokenize(prompt))

        with concurrent.futures.ThreadPoolExecutor(max_workers=concurrency) as executor:
            futures = [
                executor.submit(request, client, prompt, args.max_tokens)
                for _ in range(concurrency)
            ]
            concurrent.futures.wait(futures)
            for future in futures:
                result = future.result()
                generated_tokens = len(tokenizer.tokenize(result["output"]))
                results.append(
                    {
                        "concurrency": concurrency,
                        "i": i,
                        "prompt_size": prompt_size,
                        "generated_tokens": generated_tokens,
                        "generation_time": result["generation_time"],
                        "total_time": result["total_time"],
                    }
                )

df = pd.DataFrame(results)
df.drop(labels=df.query("generated_tokens < 100").index, inplace=True)
df["generated_tps"] = df["generated_tokens"] / df["generation_time"]
df["total_tps"] = (df["prompt_size"] + df["generated_tokens"]) / df["total_time"]
df_results = df.groupby("concurrency")[["generated_tps", "total_tps"]].agg(
    ["mean", "median", "std"]
)
print(df_results)

df_grouped = df.groupby(["concurrency", "i"])[["prompt_size", "generated_tokens"]].sum()
df_grouped["system_tokens"] = df_grouped["prompt_size"] + df_grouped["generated_tokens"]
total_times = df.groupby(["concurrency", "i"])["total_time"].median()
system_tokens = df_grouped["system_tokens"] / total_times
print()
print("System tokens")
print(system_tokens.groupby("concurrency").agg(["median", "std"]))
df.to_csv(f"benchmark.csv", index=False)

fig, ax = plt.subplots()
ax.set_title(
    f"MariTalk Local (prompt_size={args.prompt_size}, max_tokens={args.max_tokens})"
)

ax.errorbar(
    df_results.index,
    df_results["generated_tps"]["median"],
    yerr=df_results["generated_tps"]["std"],
    label="generated_tps",
    color="blue",
)
ax.errorbar(
    df_results.index,
    df_results["total_tps"]["median"],
    yerr=df_results["total_tps"]["std"],
    label="total_tps",
    color="green",
)
ax.errorbar(
    system_tokens.groupby("concurrency").median().index,
    system_tokens.groupby("concurrency").median(),
    yerr=system_tokens.groupby("concurrency").std(),
    label="system_tps",
    color="red",
)

ax.legend(loc="lower right")
ax.set_xticks(system_tokens.groupby("concurrency").median().index)
ax.set_xlabel("Concurrency")
ax.set_ylabel("Tokens/s")

plt.tight_layout()
plt.show()
plt.savefig(f"benchmark.png", bbox_inches="tight")



import os
import re
import csv
import sys
import time
import atexit
import threading
import subprocess
from tqdm import tqdm
from pathlib import Path
from typing import List, Dict, Optional, Union, Optional
import requests
from requests.exceptions import ConnectionError


def check_gpu():
    try:
        result = subprocess.run(
            [
                "nvidia-smi",
                "--query-gpu=name,compute_cap",
                "--format=csv",
            ],
            capture_output=True,
            text=True,
            check=True,
        )
        reader = csv.reader(result.stdout.strip().split("\n"))
        headers = next(reader)
        rows = list(reader)

        for row in rows:
            gpu_name, compute_cap = row
            if float(compute_cap) >= 8.0:
                return True

        raise Exception(
            f"The detected device is not supported: {gpu_name}. We currently support Nvidia GPUs with compute capability >= 8.0."
        )
    except subprocess.CalledProcessError as e:
        raise Exception(f"Error executing command: {e}")
    except FileNotFoundError as e:
        raise Exception(
            "Nvidia-SMI is not installed. Please install the Nvidia driver and the CUDA toolkit."
        )


def find_libs():
    versions = {
        "cuda_version": None,
    }

    try:
        output = subprocess.run(
            ["nvidia-smi"],
            stdout=subprocess.PIPE,
        ).stdout.decode("utf-8")
        cuda_version_match = re.search(r"CUDA Version: (\d+\.\d+)", output)

        if not cuda_version_match:
            raise Exception(
                """Could not automatically detect the CUDA version. Verify the CUDA Toolkit installation or set the `cuda_version` parameter manually. For example:

```
model.start_server("<YOUR LICENSE>", cuda_version="12.3")
```

To install the CUDA Toolkit, please refer to: https://developer.nvidia.com/cuda-downloads"""
            )

        versions["cuda_version"] = cuda_version_match.group(1)
    except subprocess.CalledProcessError as e:
        raise Exception(f"Error executing command: {e}")
    except FileNotFoundError as e:
        raise Exception(
            "Nvidia-SMI is not installed. Please install the Nvidia driver and the CUDA toolkit."
        )

    return versions


def download(license: str, bin_path: str, dependencies: Dict[str, int]):
    download_url = "https://functions.maritaca.ai/local/download"
    response = requests.post(
        download_url,
        json={
            "license": license,
            "cuda_version": dependencies["cuda_version"],
        },
    )
    if not response.ok:
        raise Exception(response.text)

    result = response.json()

    if "presigned_url" not in result:
        raise ValueError(f"Failed to validate license ({license}): {result}")

    file_url = result["presigned_url"]
    model_size = result["model_size"]

    print(f"Downloading MariTalk-{model_size} (path: {bin_path})...")

    try:
        with requests.get(file_url, stream=True) as response:
            response.raise_for_status()
            file_size = int(response.headers.get("content-length", 0))
            if file_size > 0:
                progress_bar = tqdm(
                    total=file_size,
                    unit="B",
                    unit_scale=True,
                    desc=bin_path,
                )
                with open(bin_path, "wb") as out:
                    for chunk in response.iter_content(chunk_size=16384):
                        out.write(chunk)
                        progress_bar.update(len(chunk))

                os.chmod(bin_path, 0o744)
            else:
                raise Exception(
                    f"Invalid response from the server while downloading: {response.text}"
                )
    except requests.exceptions.RequestException as e:
        raise Exception(f"Error downloading MariTalk binary: {e}")


def _get_total_mem():
    try:
        output = subprocess.check_output(["free", "-h"], text=True)
        for line in output.splitlines():
            if line.startswith("Mem:"):
                mem_info = line.split()[1]
                return _convert_to_gb(mem_info)
        return None
    except (subprocess.CalledProcessError, FileNotFoundError):
        return None


def _convert_to_gb(mem_str):
    """
    Convert memory string from `free -h` (like 251Gi) to gigabytes as a float.
    """
    match = re.match(r"([0-9.]+)([KMGTPE]i)", mem_str)
    if match:
        value, unit = match.groups()
        unit_factor = {"Mi": 1 / 1024, "Gi": 1, "Ti": 1024}
        return float(value) * unit_factor[unit]
    return None


def _get_file_size(file_path):
    try:
        file_size_bytes = os.path.getsize(file_path)
        return file_size_bytes / (1024**3)
    except Exception:
        return None


def start_server(
    license: str,
    bin_path: str = "~/bin/maritalk",
    cuda_version: Optional[int] = None,
    port: int = 9000,
):
    bin_path = os.path.expanduser(bin_path)
    if not os.path.exists(bin_path):
        if not cuda_version:
            check_gpu()
        detected_versions = find_libs()

        dependencies = {
            "cuda_version": cuda_version or detected_versions["cuda_version"]
        }

        if dependencies["cuda_version"] is None:
            raise Exception(
                "No libcublas.so found. cuBLAS v11 or v12 is required to run MariTalk. You can manually set the version using the `cuda_version` argument."
            )

        bin_folder = os.path.dirname(bin_path)
        if bin_folder:
            os.makedirs(bin_folder, exist_ok=True)
        download(license, bin_path, dependencies)

    bin_size = _get_file_size(bin_path)

    if bin_size:
        min_memory = 20 if bin_size < 20 else 70
        memory_available = _get_total_mem()
        if memory_available and memory_available < min_memory:
            print(
                "WARNING: Verify that there is enough memory to load the model (at least 20GB for the small version and 70GB for the medium version)."
            )

    env = os.environ.copy()
    env["DISABLE_LOADING"] = "1"
    args = [bin_path, "--license", license, "--port", str(port)]
    return subprocess.Popen(
        args,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        env=env,
    )


class MariTalkLocal:
    def __init__(self, host: str = "localhost", port: int = 9000):
        self.api_url = f"http://{host}:{port}"
        """@private"""
        self.port = port
        """@private"""
        self.process = None
        """@private"""
        self.loading = False
        """@private"""
        self.loaded = False
        """@private"""

    def start_server(
        self,
        license: str,
        bin_path: str = "~/bin/maritalk",
        cuda_version: Optional[int] = None,
        verbose: str = True,
    ):
        if self.loaded:
            return

        self.loading = True
        if verbose:
            print(f"Starting MariTalk Local API at http://localhost:{self.port}/")
            print(
                "This process can take a few minutes (up to 10 minutes for the small version, depending on the hardware)."
            )
            loading_thread = threading.Thread(target=self._show_loading)
            loading_thread.start()
        self.process = start_server(license, bin_path, cuda_version, self.port)

        while True:
            try:
                if self.process.poll() is not None:
                    output, _ = self.process.communicate()
                    output = output.decode("utf-8")
                    raise Exception(
                        f"Failed to start process.\nOutput: {output}\nTry to run it manually: `{' '.join(self.process.args)}`"
                    )

                response = self.status()
                if response and response["status"] == "loading":
                    continue
                break
            except ConnectionError as ex:
                time.sleep(1)

        self.loading = False
        self.loaded = True

        if verbose:
            loading_thread.join()
            print()

        def terminate():
            print("Stopping MariTalk...")
            self.stop_server()

        atexit.register(terminate)

    def _show_loading(self):
        spinner = ["⠋", "⠙", "⠚", "⠞", "⠖", "⠦", "⠴", "⠲", "⠳", "⠓"]
        spinner_index = 0
        start_time = time.time()

        try:
            while self.loading:
                current_time = time.time()
                elapsed_time = int(current_time - start_time)
                minutes, seconds = divmod(elapsed_time, 60)

                output = (
                    f"\rLoading... {spinner[spinner_index]} ({minutes}min:{seconds:02d}s)"
                )
                sys.stdout.write(output)
                sys.stdout.flush()
                spinner_index = (spinner_index + 1) % len(
                    spinner
                )
                time.sleep(0.1)
        except KeyboardInterrupt:
            sys.stdout.flush()

    def stop_server(self):
        if not self.process:
            print("No process attached to this client!")
            return
        self.process.terminate()
        self.process = None
        self.loaded = False

    def status(self):
        response = requests.get(self.api_url)
        return response.json()

    def generate_raw(
        self,
        prompt: str,
        temperature: float = 0.7,
        top_p: float = 0.95,
        max_tokens: int = 512,
        do_sample: bool = True,
        stop_sequences: List[str] = [],
        seed: Optional[int] = None,
    ):
        """
        Generate a completion for a given prompt.

        Args:
            prompt (`str`):
                The initial prompt to generate completions for. It can be the beginning of a text to be completed or an instruction for performing a task.
            temperature (`float`, *optional*, defaults to `0.7`):
                The sampling temperature for the next token probability. Higher values generate more random texts, while lower values will make it more deterministic.
            top_p (`float`, *optional*, defaults to `0.95`):
                The top probability mass to use on nucleus sampling. Read more at: https://arxiv.org/abs/1904.09751.
            max_tokens (`int`, *optional*, defaults to `512`):
                Maximum number of tokens to generate.
            do_sample (`bool`, *optional*, defaults to `True`):
                Whether to use sampling or not. `True` value means non-deterministic generations using sampling parameters and `False` value means deterministic generation using greedy decoding.
            stop_sequences (`List[str]`, *optional*):
                A list of sequences to stop the generation process.
            seed (`int`, *optional*, defaults to `None`):
                Seed to use during the random sampling process to make it reproducible.
        """

        body = {
            "prompt": prompt,
            "do_sample": do_sample,
            "temperature": temperature,
            "top_p": top_p,
            "max_tokens": max_tokens,
            "stop_sequences": stop_sequences,
            "seed": seed,
        }

        response = requests.post(f"{self.api_url}/generate", json=body)

        if response.ok:
            return response.json()

        # Return the error message in case of a failed request
        if response.headers.get("content-type") == "application/json":
            return response.json()

        # Raise for any other failed responses
        response.raise_for_status()

    def generate(
        self,
        messages: Union[str, List[Dict[str, str]]],
        temperature: float = 0.7,
        top_p: float = 0.95,
        max_tokens: int = 512,
        do_sample: bool = True,
        stop_sequences: List[str] = [],
        seed: Optional[int] = None,
    ):
        """
        Generate a completion for a given prompt.

        Args:
            messages (`Union[str, List[Dict[str, str]]]`):
                A string with one message or a list where each item should be a dictionary containing the keys `role` and `content`. For example:
                ```
                messages = [
                    {"role": "user", "content": "bom dia, esta é a mensagem do usuario"},
                    {"role": "assistant", "content": "bom dia, esta é a resposta do assistente"},
                    {"role": "user", "content": "Você pode me falar quanto é 25 + 27?"},
                ]
                ```
            temperature (`float`, *optional*, defaults to `0.7`):
                The sampling temperature for the next token probability. Higher values generate more random texts, while lower values will make it more deterministic.
            top_p (`float`, *optional*, defaults to `0.95`):
                The top probability mass to use on nucleus sampling. Read more at: https://arxiv.org/abs/1904.09751.
            max_tokens (`int`, *optional*, defaults to `512`):
                Maximum number of tokens to generate.
            do_sample (`bool`, *optional*, defaults to `True`):
                Whether to use sampling or not. `True` value means non-deterministic generations using sampling parameters and `False` value means deterministic generation using greedy decoding.
            stop_sequences (`List[str]`, *optional*):
                A list of sequences to stop the generation process.
            seed (`int`, *optional*, defaults to `None`):
                Seed to use during the random sampling process to make it reproducible.
        """

        if not isinstance(messages, str) and not isinstance(messages, list):
            raise TypeError(
                "You should pass a string or a list of messages as argument."
            )

        if isinstance(messages, str):
            messages = [{"role": "user", "content": messages}]

        body = {
            "messages": messages,
            "do_sample": do_sample,
            "temperature": temperature,
            "top_p": top_p,
            "max_tokens": max_tokens,
            "stop_sequences": stop_sequences,
            "seed": seed,
        }

        response = requests.post(f"{self.api_url}/chat/generate", json=body)

        if response.ok:
            return response.json()

        # Return the error message in case of a failed request
        if response.headers.get("content-type") == "application/json":
            return response.json()

        # Raise for any other failed responses
        response.raise_for_status()

    def generate_chat(self, *args, **kwargs):
        raise Exception(
            "This method was changed, please use `generate` for chat messages or `generate_raw` for raw few-shot examples instead."
        )
